#ifdef LOG_LOCAL_LEVEL
#undef LOG_LOCAL_LEVEL
#define LOG_LOCAL_LEVEL IOTMODULE_LOG_LEVEL
#else
#define LOG_LOCAL_LEVEL IOTMODULE_LOG_LEVEL
#endif

static const char* pTAG = "";

//================================================================================================
//                                  cIotModuleAdc body
//================================================================================================

cIotModuleAdc::cIotModuleAdc()
{
}

void cIotModuleAdc::Initialize(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  SpiActive = false;
  ISR_Active = false;
  MasterPinConfigured = false;
  ContinuousConversion_ContRead = true;
  //  reset conversion data buffer
  for (int i = 0; i < IOTMODULE_ADC24_PINS; i++)
  {
    ISR_Channel_uint32_Value[i] = 0;
  }
  //  reset ISR counter
  ISR_Counter = 0;
}

void cIotModuleAdc::SetSPI_Control_Pin(uint8_t Type, gpio_num_t Name, gpio_pull_mode_t ResetMode)
{
  switch(Type)
  {
    case AD7124_MOSI:
      SPI_MOSI_Name = Name;
      SPI_MOSI_ResetMode = ResetMode;
      break;
    case AD7124_MISO:
      SPI_MISO_Name = Name;
      SPI_MISO_ResetMode = ResetMode;
      break;
    case AD7124_CLK:
      SPI_CLK_Name = Name;
      SPI_CLK_ResetMode = ResetMode;
      break;
  }
}

void cIotModuleAdc::ShutDown(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - ShutDown called", __FILENAME__, __LINE__);
  ISR_Detach();
  SPI_Detach();
  //  reset conversion data buffer
  for (int i = 0; i < IOTMODULE_ADC24_PINS; i++)
  {
    ISR_Channel_uint32_Value[i] = 0;
  }

  ESP_ERROR_CHECK(gpio_reset_pin(SPI_MOSI_Name));
  ESP_ERROR_CHECK(gpio_set_direction(SPI_MOSI_Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
  ESP_ERROR_CHECK(gpio_set_pull_mode(SPI_MOSI_Name, SPI_MOSI_ResetMode));  
  
  ESP_ERROR_CHECK(gpio_reset_pin(SPI_MISO_Name));
  ESP_ERROR_CHECK(gpio_set_direction(SPI_MISO_Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
  ESP_ERROR_CHECK(gpio_set_pull_mode(SPI_MISO_Name, SPI_MISO_ResetMode));  

  ESP_ERROR_CHECK(gpio_reset_pin(SPI_CLK_Name));
  ESP_ERROR_CHECK(gpio_set_direction(SPI_CLK_Name, (gpio_mode_t)GPIO_MODE_DEF_INPUT));
  ESP_ERROR_CHECK(gpio_set_pull_mode(SPI_CLK_Name, SPI_CLK_ResetMode));  

  //  reset ISR counter
  ISR_Counter = 0;
  MasterPinConfigured = false;
}

void cIotModuleAdc::SetParam(const uint16_t Sps, const bool MultipleSetups, const bool FiltersSingleCycle, const uint8_t PreFilterType,
                             const uint8_t PostFilterType, const uint16_t ActiveChannels, const uint16_t BiasedChannels, const uint16_t DiffChannels,
                             const uint16_t PseudoChannels, const uint8_t ChannelsGains0, const uint8_t ChannelsGains1, const uint8_t ChannelsGains2)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - SetParam(Adc) called", __FILENAME__, __LINE__);

  //  if ADC24 configured and master pin configured
  if (IOTMODULE_ADC24_PIN_MASTER == 1 && MasterPinConfigured)
  {
    _Sps = Sps;
    _MultipleSetups = MultipleSetups;
    _FiltersSingleCycle = FiltersSingleCycle;
    _PreFilterType = PreFilterType;
    _PostFilterType = PostFilterType;
    _ActiveChannels = ActiveChannels;
    _BiasedChannels = BiasedChannels;
    _DiffChannels =   DiffChannels;
    _PseudoChannels = PseudoChannels;
    _ChannelsGains0 = ChannelsGains0;
    _ChannelsGains1 = ChannelsGains1;
    _ChannelsGains2 = ChannelsGains2;
    //  count active channels
    uint8_t Index = 15, Count = 0;
    do
    {
      if (ActiveChannels & (1 << Index))
      {
        Count++;
      }
    }
    while (Index-- != 0);
    //  stima durata interrupt con lettura in polling
    //  X = 2 uS latenza interrupt
    //  Y = 6 uS latenza funzione lettura SPI
    //  X + Y + 2 @16MHz
    //  X + Y + 4 @8MHz
    //  X + Y + 8 @4MHz
    //  X + Y + 16 @2MHz
    //  X + Y + 32 @1MHz
    uint32_t SPI_Speed = IOTMODULE_ADC24_COMMUNICATION_SPEED;
    //  if value out of range
    if (SPI_Speed == 0 || SPI_Speed > 16)
    {
      SPI_Speed = 16;
    }
    //  if only one channel defined and speed too low
    if (Count == 1 && SPI_Speed == 1)
    {
      SPI_Speed = 2;
    }
    //  detach GPIO ISR
    ISR_Detach();
    //  detach SPI
    SPI_Detach();
    //  attach SPI with new speed
    SPI_Attach(SPI_CLK_Name, SPI_MISO_Name, SPI_MOSI_Name, SPI_Speed * 1000 * 1000);
    //  reset AD7124
    Reset();
    //  set idle mode
    WaitRegistersWritableBySPI();
    Init_Idle();
    //  power on reset status flag must be set after reset
    //  if not, the chip may be disconnected
    if (!(ReadStatusRegister() & AD7124_STATUS_REGISTER_POR_FLAG))
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - AD7124 not in reset status", __FILENAME__, __LINE__);
      //  ignore chip
      Count = 0;
    }
    //  read ID
    Index = ReadID_Register();
    //  ID must be different from 0x00 or 0xff
    //  if not, the chip may be disconnected
    if (Index == 0x00 || Index == 0xff)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - AD7124 ID not recognized", __FILENAME__, __LINE__);
      //  ignore chip
      Count = 0;
    }
    //  if active channels
    if (Count > 0)
    {
      //  program AD7124 registers
      ProgramAllRegisters();
      //  any error?
      if (ReadStatusRegister() & AD7124_STATUS_REGISTER_ERROR_FLAG)
      {
        uint32_t ErrorRegister = ReadReg24(AD7124_ERROR_REGISTER);
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - AD7124_STATUS_REGISTER_ERROR_FLAG is set, Error : %08x", __FILENAME__, __LINE__, ErrorRegister);
      }
      //  if continuous conversion and continuous read
      if (ContinuousConversion_ContRead)
      {
        //  continuous conversion and continuous read
        WaitRegistersWritableBySPI();
        Init_ContinuousConversion_ContRead();
      }
      else
      {
        //  only continuous conversion
        WaitRegistersWritableBySPI();
        Init_ContinuousConversion();
      }
      //  attach GPIO ISR
      ISR_Attach();
      
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - AD7124 ID register : 0x%02X", __FILENAME__, __LINE__, Index);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SPI communication speed : %d MHz", __FILENAME__, __LINE__, SPI_Speed);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - programmed channels : %d", __FILENAME__, __LINE__, Count);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - programmed SPS : %d", __FILENAME__, __LINE__, Sps);
    }
  }
  //  if ADC24 not configured
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - configuration function called but IOTMODULE_ADC24_PIN_MASTER is not configured", __FILENAME__, __LINE__);
  }
}

void cIotModuleAdc::ProgramAllRegisters(void)
{
  uint8_t i;
  uint16_t ui16;
  uint32_t ui24;
  uint32_t ChannelGains24;

  ChannelGains24 = ((uint32_t)_ChannelsGains2 << 16);
  ChannelGains24 += ((uint32_t)_ChannelsGains1 << 8);
  ChannelGains24 += ((uint32_t)_ChannelsGains0);

  WaitRegistersWritableBySPI();
  WriteReg16(AD7124_IOCon2, _BiasedChannels);

  //  Prepare Active / Diff / Pseudo / SingleEnded
  for (i = 0; i < 16; i++)
  {
    ui16 = 0;
    if ((_ActiveChannels >> i) & 1)
    {
      ui16 |= AD7124_CH_MAP_REG_CH_ENABLE;
      if (_MultipleSetups)
      {
        // CH0/1 = Setup0 ... CH14/15 = Setup7
        ui16 |= ((i / 2) << 12);
      }
      // AINP = AIN_i << 5
      ui16 |= (i << 5);
      if ((_DiffChannels >> i) & 1)
      {
        // AINM = AIN_i + 1
        ui16 |= (i + 1);
      }
      else if ((_PseudoChannels >> i) & 1)
      {
        // AINM = 15
        ui16 |= 15;
      }
      else
      {
        // AINM = AVss
        ui16 |= 0b10001;
      }
    }
    WaitRegistersWritableBySPI();
    WriteReg16(AD7124_Channel_0 + i, ui16);
  }

  //  Bipolar or SingleEnded and Gains
  for (i = 0; i < 8; i++)
  {
    ui16 = AD7124_CFG_REG_AIN_BUFP |
           AD7124_CFG_REG_AINN_BUFM |
           AD7124_CFG_REG_REF_SEL(0b11);

    if ( ((_DiffChannels >> (i * 2)) & 0b11) ||
         ( (_PseudoChannels >> (i * 2)) & 0b11) )
      ui16 |= AD7124_CFG_REG_BIPOLAR;

    ui16 += (ChannelGains24 >> (i * 3)) & 0b111;

    WaitRegistersWritableBySPI();
    WriteReg16(AD7124_Config_0 + i, ui16);
  }

  //  PreFilter, PostFilter and Sps
  for (i = 0; i < 8; i++)
  {
    ui24 = 0;
    ui24 += ((uint32_t)_PreFilterType << 21);
    ui24 += ((uint32_t)_PostFilterType << 17);
    if (_FiltersSingleCycle)
      ui24 |= AD7124_FILT_REG_SINGLE_CYCLE;
    if (_Sps < 10) _Sps = 10;
    ui24 += ((uint32_t)19200 / _Sps);
    WaitRegistersWritableBySPI();
    WriteReg24(AD7124_Filter_0 + i, ui24);
  }
}

void cIotModuleAdc::ConfigureMasterPin(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - ConfigureMasterPin called", __FILENAME__, __LINE__);
  MasterPinConfigured = true;
}

void cIotModuleAdc::PreFuota(void)
{
  //  Stop any interrupt activity
  ISR_Detach();
  SPI_Detach();
}

void cIotModuleAdc::ISR_Attach(void)
{
  if (!ISR_Active)
  {
    ESP_ERROR_CHECK(gpio_isr_handler_add(SPI_MISO_Name, SPI_ISR_Handler, this));
    ESP_ERROR_CHECK(gpio_set_intr_type(SPI_MISO_Name, GPIO_INTR_NEGEDGE));
    //  ESP_ERROR_CHECK(gpio_intr_enable((SPI_MISO_Name));
    ISR_Active = true;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - ISR_Active : true", __FILENAME__, __LINE__);
  }
}

void cIotModuleAdc::ISR_Detach(void)
{
  if (ISR_Active)
  {
    //  ESP_ERROR_CHECK(gpio_intr_disable(SPI_MISO_Name));
    ESP_ERROR_CHECK(gpio_isr_handler_remove(SPI_MISO_Name));
    ISR_Active = false;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - ISR_Active : false", __FILENAME__, __LINE__);
  }
}

void cIotModuleAdc::SPI_Attach(const uint8_t PinNumClk, const uint8_t PinNumMiso, const uint8_t PinNumMosi, const uint32_t ClockSpeedHz)
{
  if (!SpiActive)
  {
    //  Initialize the SPI bus
    spi_bus_config_t buscfg;
    memset(&buscfg, 0, sizeof(spi_bus_config_t));
    buscfg.miso_io_num = PinNumMiso;
    buscfg.mosi_io_num = PinNumMosi;
    buscfg.sclk_io_num = PinNumClk;
    buscfg.quadwp_io_num = -1;
    buscfg.quadhd_io_num = -1;
    buscfg.max_transfer_sz = 8;
    ESP_ERROR_CHECK(spi_bus_initialize(HSPI_HOST, &buscfg, 0));

    //  Initialize the SPI device
    spi_device_interface_config_t devcfg;
    memset(&devcfg, 0, sizeof(spi_device_interface_config_t));
    devcfg.command_bits = 8;
    devcfg.address_bits = 0;
    devcfg.dummy_bits = 0;
    //  MUST BE MODE 3 for continuous conversion & continuous read
    //  may be mode 0 or 3 for continuous conversion without continuous read
    devcfg.mode = 3;
    devcfg.duty_cycle_pos = 0;
    devcfg.cs_ena_pretrans = 0;
    devcfg.cs_ena_posttrans = 0;
    //  SPI clock frequency
    devcfg.clock_speed_hz = ClockSpeedHz;
    devcfg.input_delay_ns = 0;
    //  CS pin not used
    devcfg.spics_io_num = -1;
    devcfg.queue_size = 1;
    devcfg.pre_cb = nullptr;
    devcfg.post_cb = nullptr;
    //  Bitwise OR of SPI_DEVICE_* flags
    devcfg.flags = 0;
    ESP_ERROR_CHECK(spi_bus_add_device(HSPI_HOST, &devcfg, &hSpi));

    SpiActive = true;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SpiActive : true", __FILENAME__, __LINE__);
  }
}

void cIotModuleAdc::SPI_Detach()
{
  if (SpiActive)
  {
    ESP_ERROR_CHECK(spi_bus_remove_device(hSpi));
    ESP_ERROR_CHECK(spi_bus_free(HSPI_HOST));
    SpiActive = false;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SpiActive : false", __FILENAME__, __LINE__);
  }
}

void cIotModuleAdc::Reset(void)
{
  if (SpiActive)
  {
    uint8_t resetbytes[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
    spi_transaction_ext_t trans;
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    //  SPI_TRANS_VARIABLE_CMD flag is set to override default command_bits value defined in spi_device_interface_config_t
    //  if SPI_TRANS_VARIABLE_CMD is set (here it is) command_bits field in spi_transaction_ext_t (set with 0) is taken in account
    //  and the 8 bit cmd field is not used in transaction
    trans.base.flags = SPI_TRANS_VARIABLE_CMD;
    trans.base.length = 8 * 8;
    trans.base.tx_buffer = resetbytes;
    ESP_ERROR_CHECK(spi_device_polling_transmit(hSpi, (spi_transaction_t*)&trans));
    //  reset conversion data buffer
    for (int i = 0; i < IOTMODULE_ADC24_PINS; i++)
    {
      ISR_Channel_uint32_Value[i] = 0;
    }
    //  delay must be at least 90 * tMCLK, is 90 * 614.4 kHz, is 90 * 1,62 uS, is 142,5 uS
    delay(1);
  }
}

uint8_t cIotModuleAdc::GetISR_Counter()
{
  return ISR_Counter;
}

uint32_t cIotModuleAdc::GetISR_Channel_uint32_Value(const uint8_t Index)
{
  //  read conversion value relative to the channel indexed
  return ISR_Channel_uint32_Value[Index & (IOTMODULE_ADC24_PINS - 1)];
}

void IRAM_ATTR cIotModuleAdc::SPI_ISR_Handler(void* arg)
{
  uint32_t r;
  spi_transaction_ext_t trans;
  cIotModuleAdc *pIotModuleAdc = (cIotModuleAdc *)arg;

  if (pIotModuleAdc->SpiActive)
  {
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    //  in both modes, if adc24 power is lost, we are no more able to read data correctly
    //  because in default configuration AD7124_ADC_CTRL_REG_DATA_STATUS bit is not set
    //  and chip do not feeds 8 bit status register data after 24 bit channel conversion data
    //  until the chip is reprogrammed again. for this reason in the ISR
    //  AD7124_STATUS_REGISTER_ERROR_FLAG is not monitored

    //  4 bytes / 32 bits transfer (data[4]))
    if (pIotModuleAdc->ContinuousConversion_ContRead)
    {
      //  SPI_TRANS_VARIABLE_CMD flag is set to override default command_bits value defined in spi_device_interface_config_t
      //  if SPI_TRANS_VARIABLE_CMD is set (here it is) command_bits field in spi_transaction_ext_t (set with 0) is taken in account
      //  and the 8 bit cmd field is not used in transaction
      trans.base.flags = SPI_TRANS_USE_RXDATA | SPI_TRANS_USE_TXDATA | SPI_TRANS_VARIABLE_CMD;
      trans.base.length = 4 * 8;

      ESP_ERROR_CHECK(spi_device_polling_transmit(pIotModuleAdc->hSpi, (spi_transaction_t*)&trans));

      r = (uint32_t)trans.base.rx_data[0] << 24;
      r |= (uint32_t)trans.base.rx_data[1] << 16;
      r |= (uint32_t)trans.base.rx_data[2] << 8;
      r |= (uint32_t)trans.base.rx_data[3];
    }
    else
      //  5 bytes / 40 bits transfer (cmd + data[4])
    {
      //  SPI_TRANS_VARIABLE_CMD flag is set to override default command_bits value defined in spi_device_interface_config_t
      //  if SPI_TRANS_VARIABLE_CMD is not set (default) command_bits field in spi_transaction_ext_t is ignored
      //  and the 8 bit cmd field is sent in trasnsaction
      trans.base.cmd = AD7124_COMMS_WRITE_FLAG + (AD7124_DATA_REGISTER & 0x3F);
      trans.base.flags |= SPI_TRANS_USE_TXDATA | SPI_TRANS_USE_RXDATA;
      trans.base.length = 4 * 8;

      ESP_ERROR_CHECK(spi_device_polling_transmit(pIotModuleAdc->hSpi, (spi_transaction_t*)&trans));

      r = (uint32_t)trans.base.rx_data[0] << 24;
      r |= (uint32_t)trans.base.rx_data[1] << 16;
      r |= (uint32_t)trans.base.rx_data[2] << 8;
      r |= (uint32_t)trans.base.rx_data[3];
    }
    //  set value
    pIotModuleAdc->ISR_Channel_uint32_Value[r & (IOTMODULE_ADC24_PINS - 1)] = r >> 8;
    //  if no error
    if (!(r & AD7124_STATUS_REGISTER_ERROR_FLAG))
    {
      //  increment activity
      pIotModuleAdc->ISR_Counter++;
    }
  }
}

void cIotModuleAdc::WriteReg16(const uint8_t reg, const uint16_t data16)
{
  spi_transaction_ext_t trans;
  if (SpiActive)
  {
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    trans.base.cmd = (reg & 0x3F);

    trans.base.length = 2 * 8;
    trans.base.flags |= SPI_TRANS_USE_TXDATA | SPI_TRANS_USE_RXDATA;
    trans.base.tx_data[0] = data16 >> 8;
    trans.base.tx_data[1] = data16 & 0xff;
    ESP_ERROR_CHECK(spi_device_polling_transmit(hSpi, (spi_transaction_t*)&trans));
  }
}

void cIotModuleAdc::WriteReg24(const uint8_t reg, const uint32_t data24)
{
  spi_transaction_ext_t trans;
  if (SpiActive)
  {
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    trans.base.cmd = (reg & 0x3F);

    trans.base.length = 3 * 8;
    trans.base.flags |= SPI_TRANS_USE_TXDATA | SPI_TRANS_USE_RXDATA;
    trans.base.tx_data[0] = data24 >> 16;
    trans.base.tx_data[1] = data24 >> 8;
    trans.base.tx_data[2] = data24 & 0xff;
    ESP_ERROR_CHECK(spi_device_polling_transmit(hSpi, (spi_transaction_t*)&trans));
  }
}

uint8_t cIotModuleAdc::ReadReg8(const uint8_t reg)
{
  spi_transaction_ext_t trans;
  uint8_t retval = 0;
  if (SpiActive)
  {
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    trans.base.cmd = AD7124_COMMS_WRITE_FLAG + (reg & 0x3F);

    trans.base.length = 1 * 8;
    trans.base.flags |= SPI_TRANS_USE_TXDATA | SPI_TRANS_USE_RXDATA;
    ESP_ERROR_CHECK(spi_device_polling_transmit(hSpi, (spi_transaction_t*)&trans));

    retval = (uint32_t)trans.base.rx_data[0];
  }
  return retval;
}

uint32_t cIotModuleAdc::ReadReg24(const uint8_t reg)
{
  spi_transaction_ext_t trans;
  uint32_t retval = 0;
  if (SpiActive)
  {
    memset(&trans, 0, sizeof(spi_transaction_ext_t));

    trans.base.cmd = AD7124_COMMS_WRITE_FLAG + (reg & 0x3F);

    trans.base.length = 3 * 8;
    trans.base.flags |= SPI_TRANS_USE_TXDATA | SPI_TRANS_USE_RXDATA;
    ESP_ERROR_CHECK(spi_device_polling_transmit(hSpi, (spi_transaction_t*)&trans));

    retval = (uint32_t)(trans.base.rx_data[0] << 16);
    retval |= (uint32_t)(trans.base.rx_data[1] << 8);
    retval |= (uint32_t)(trans.base.rx_data[2]);
  }
  return retval;
}

uint16_t cIotModuleAdc::GetActiveChannels(void)
{
  return _ActiveChannels;
}

bool cIotModuleAdc::GetMasterPinConfigured(void)
{
  return MasterPinConfigured;
}

void cIotModuleAdc::SetFixedParam(void)
{
  //  early test configuration values
  _Sps = 50;
  _MultipleSetups = true;
  _FiltersSingleCycle = false;
  _PreFilterType = 2;
  _PostFilterType = 3;
  _ActiveChannels = 0b01010101;
  _BiasedChannels = 0;
  _DiffChannels =   0b11111111;
  _PseudoChannels = 0;
  _ChannelsGains0 = 0;
  _ChannelsGains1 = 0;
  _ChannelsGains2 = 0;
}

void cIotModuleAdc::Init_ContinuousConversion(void)
{
  WriteReg16(AD7124_ADC_CONTROL_REGISTER,
             AD7124_ADC_CTRL_REG_DOUT_RDY_DEL |     //  ready delayed
             AD7124_ADC_CTRL_REG_DATA_STATUS |      //  conversion data + status register
             AD7124_ADC_CTRL_REG_REF_EN |           //  reference enabled
             AD7124_ADC_CTRL_REG_POWER_MODE(0b11) | //  full power
             AD7124_ADC_CTRL_REG_MODE(0b0000) |     //  continuous conversion
             AD7124_ADC_CTRL_REG_CLK_SEL(0b00));    //  614.4 kHz
}

void cIotModuleAdc::Init_ContinuousConversion_ContRead(void)
{
  WriteReg16(AD7124_ADC_CONTROL_REGISTER,
             AD7124_ADC_CTRL_REG_DOUT_RDY_DEL |     //  ready delayed
             AD7124_ADC_CTRL_REG_CONT_READ |        //  continuous read
             AD7124_ADC_CTRL_REG_DATA_STATUS |      //  conversion data + status register
             AD7124_ADC_CTRL_REG_REF_EN |           //  reference enabled
             AD7124_ADC_CTRL_REG_POWER_MODE(0b11) | //  full power
             AD7124_ADC_CTRL_REG_MODE(0b0000) |     //  continuous conversion
             AD7124_ADC_CTRL_REG_CLK_SEL(0b00));    //  614.4 kHz
}

void cIotModuleAdc::Init_Idle(void)
{
  WriteReg16(AD7124_ADC_CONTROL_REGISTER,
             AD7124_ADC_CTRL_REG_DOUT_RDY_DEL |     //  ready delayed
             AD7124_ADC_CTRL_REG_DATA_STATUS |      //  conversion data + status register
             AD7124_ADC_CTRL_REG_REF_EN |           //  reference enabled
             AD7124_ADC_CTRL_REG_POWER_MODE(0b11) | //  full power
             AD7124_ADC_CTRL_REG_MODE(0b0100) |     //  idle
             AD7124_ADC_CTRL_REG_CLK_SEL(0b00));    //  614.4 kHz
}

uint8_t cIotModuleAdc::ReadStatusRegister(void)
{
  //  read status register
  return ReadReg8(AD7124_STATUS_REGISTER);
}

uint8_t cIotModuleAdc::ReadID_Register(void)
{
  //  read ID register
  return ReadReg8(AD7124_ID_REGISTER);
}

void cIotModuleAdc::WaitRegistersWritableBySPI(void)
{
  //  from AD7124 manual 
  //  SPI_IGNORE ERROR
  //  At certain times, the on-chip registers are not accessible.
  //  For example, during power-up, the on-chip registers are set to their default values.
  //  The user must wait until this operation is complete before writing to registers.
  //  Also when offset or gain calibrations are being performed, registers cannot be accessed.
  //  The SPI_IGNORE_ERR bit in the error register indicates when the onchip registers cannot be written to.
  //  This diagnostic is enabled by default.
  //  The function can be disabled using the SPI_IGNORE_ERR_EN bit in the ERROR_EN register.
  //  Any write operations performed when SPI_IGNORE_ERR is enabled are ignored

  //  wait registers writable by SPI
  if(true)
  {
    uint32_t SpiIgnoreError;
    uint8_t SpiIgnoreErrorCounter = 0;
    do
    {
      SpiIgnoreError = ReadReg24(AD7124_ERROR_REGISTER) & AD7124_ERROR_REGISTER_SPI_IGNORE_ERR;
      if(SpiIgnoreError)
      {
        SpiIgnoreErrorCounter++;
      }
    }    
    while(SpiIgnoreError);
    if(SpiIgnoreErrorCounter)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - SpiIgnoreErrorCounter : %d", __FILENAME__, __LINE__, SpiIgnoreErrorCounter);
    }
  }
  else
  {
    delay(1);
  }
}

