
// ==================================================================================
//  This program reads a BNO055 sensor and sends X, Y, Z to IotHAL
//  Values are sent to IotHAL as 16 bits from 0 to 65535
// ==================================================================================
//  This version limits the angles to X/Y/Z max and min
// ==================================================================================

// ==================================================================================
//  USAGE
// ==================================================================================
//  RESET BUTTON
//   To reset you could use the reset pulse on the Arduino Board.
//   Or you could use an external pulse connected from RESET PIN to GND PIN.
//
//  CALIBRATION BUTTON
//   A button connected from PIN 18 to GND PIN. 
//
//  FLASHING LED
//   You could use the little red LED on the board.
//   Or an external LED from GND to PIN 19, via a 220 ohm resistor. 
//
//  LONG STARTUP CALIBRATION
//   1) Press the calibration button and hold it pressed.
//   2) Setup the power supply voltage, or press the reset button. 
//   3) Wait until the LED is flashing rapidly.
//   4) Release the calibration button.
//   5) Rotate sensor (SLOWLY) in some wild 3D orientation, hold it there about two seconds.
//   6) Gently rotate to some other 3D orientations, hold it there about two seconds.
//   7) Put the sensor an the table and hold it there about two seconds. 
//   8) Continue doing rotate-holds-table...  until LED stops. 
//   9) Position the sensor in the working position.
//  10) Press and release the calibration button.
//
//  FAST STARTUP CALIBRATION
//   1) Setup the power supply voltage, or press the reset button. 
//   2) Wait until the LED is flashing rapidly.
//   3) Move the sensor until LED stops. 
//   4) Position the sensor in the working position.
//   5) Press and release the calibration button.
//
//  CALIBRATION WHILE WORKING 
//  (THIS PROCEDURE IS RACCOMENDED TO OBTAIN THE MAX LONG TERM STABILITY)
//   1) Move and rotate the sensor for long time
//   2) Position the sensor in the working position.
//   3) Press and release the calibration button.
// ----------------------------------------------------------------------------------


// ==================================================================================
//  This file INO requires the following files
// ==================================================================================
// PWM.h - ATimerDefs.cpp - ATimerDefs.h - BTimerDefs.cpp - BTimerDefs.h 
// Adafruit_BNO055.cpp - Adafruit_BNO055.h - Adafruit_Sensor.h
// And the folder "utility" with: imumaths.h - matrix.h - quaternion.h - vector.h

// ==================================================================================
//  Or you could use this file INO alone, and install the following libraries
// ==================================================================================
//  Adafruit BNO055 library:
//   https://www.arduinolibraries.info/libraries/adafruit-bno055

// ==================================================================================
//   ESP32 to BNO055
// ==================================================================================
//   ESP32 Pin      BNO055  
//  ---------------------------------------------------------------------------------
//     GND          GND 
//     +5V          VIN
//     27           SDA        
//     14           SCL    
// ----------------------------------------------------------------------------------   

// ==================================================================================
//   ESP32 TO IOT-HAL
// ==================================================================================
//   ESP32 Pin     Signal type                       IotHal Pin configuration 
//  ---------------------------------------------------------------------------------
//      36          X (rotation)                      GenericIn16
//      39          Y (tilt)                          GenericIn16
//      25 (or 34)  Z (pan)                           GenericIn16
//      19          Led calibration       
//      18          Pulse to GND (LOW = calibrate) 
// ----------------------------------------------------------------------------------     

#include "IotModule\IotModule.h"
#include <Wire.h>
#include <EEPROM.h>
#include "Adafruit_Sensor.h"
#include "Adafruit_BNO055.h"
#include "utility/imumaths.h"

// ===================================
//  IotModule object, do not remove
// ===================================
cIotModule IotModule;
// ===================================

// -------------------------------------------------------- eventually modify limits
float Xmax = 90;
float Xmin = -90;
float Ymax = 90;
float Ymin = -90;
float Zmax = 90;
float Zmin = -90;

// -------------------------------------------------------- eventually modify pins
byte pinX = 36;
byte pinY = 39;
byte pinZ = 25;     // 25 on Pico - 34 on Wroom
byte pinLed = 19;
byte pinCalibrate = 18;
byte pinTest = 5;
byte pinSDA = 27;
byte pinSCL = 14;

// -------------------------------------------------------- do not modify this part
Adafruit_BNO055 bno = Adafruit_BNO055();
imu::Vector<3> euler;
float limit1 = -90;
float limit2 = 90;
float limit3 = 360;
float x = 0, y = 0, z = 0;
float calx = 0, caly = 0, calz = 0;
float oldx, oldy, oldz;
float newx, newy, newz;
float finalx, finaly, finalz;
const int k_Center = 32768;              // center for 16 bits 
const float k_MinMax = 32767.0 / 90.0;   // With 90° --> Min = 0 and Max = 65535

void setup(void)
{
  // =========================================================
  //  INIT SERIAL
  // =========================================================
  Serial.begin(115200);

  // =========================================================
  //  IotModule object initialization, do not remove
  // =========================================================
  IotModule.initialize();

  // =========================================================
  //  Init I2C communication  (SDA / SCL / FREQUENCY)
  // =========================================================
  Wire.begin(pinSDA, pinSCL);

  // =========================================================
  //  pin mode for CALIBRATE and TEST
  // =========================================================
  pinMode(pinCalibrate, INPUT_PULLUP);
  pinMode(pinLed, OUTPUT);
  pinMode(pinTest, OUTPUT);

  // =========================================================
  //  INIT BNO - default mode after begin is NDOF
  // =========================================================
  Serial.println("\nOrientation Sensor Raw Data Test"); Serial.println("");
  if(!bno.begin())
  {
    Serial.print("Ooops, no BNO055 detected ... Check your wiring or I2C ADDR!");
    while(1);
  }
  Serial.println("BNO055 connected.\n");

  // --------------------------------------- set chrystal (needs more than 300 mS)
  bno.setExtCrystalUse(true);
  delay(400);

  // --------------------------------------- test if christal is started 
  displayTemperature();

  // =========================================================
  //  CALIBRATION (if pinCalibrate = LOW)
  // =========================================================
  if (!digitalRead(pinCalibrate))
  { 
    waitCalibration();
    writeCalibrationToEeprom();
  }
  else
  {
    readCalibrationFromEeprom();
    waitCalibration();
  }
}

void loop(void)
{
  // -------------------------------------------------------------------- read euler vector (1.0 2.0 mS)
  euler = bno.getVector(Adafruit_BNO055::VECTOR_EULER);

  // -------------------------------------------------------------------- subtract calibration (0.1 mS)
  newx = euler.x() - calx;
  while (newx > 180) newx -= 360;
  while (newx < -180) newx += 360;

  newy = euler.y() - caly;
  while (newy > 180) newy -= 360;
  while (newy < -180) newy += 360;

  newz = euler.z() - calz;
  while (newz > 180) newz -= 360;
  while (newz < -180) newz += 360;

  // -------------------------------------------------------------------- recover full rotation (0.1 mS)
  x += newx - oldx;
  if (newx > limit2 && oldx < limit1) x -= limit3; 
  if (newx < limit1 && oldx > limit2) x += limit3;    
  oldx = newx;

  y += newy - oldy;
  if (newy > limit2 && oldy < limit1) y -= limit3; 
  if (newy < limit1 && oldy > limit2) y += limit3;  
  oldy = newy;

  z += newz - oldz;
  if (newz > limit2 && oldz < limit1) z -= limit3; 
  if (newz < limit1 && oldz > limit2) z += limit3;  
  oldz = newz;

  // -------------------------------------------------------------------- limit values (0.1 mS)
  finalx = x;
  finaly = y;
  finalz = z;
  if (finalx > Xmax) finalx = Xmax;
  if (finalx < Xmin) finalx = Xmin;
  if (finaly > Ymax) finaly = Ymax;
  if (finaly < Ymin) finaly = Ymin;
  if (finalz > Zmax) finalz = Zmax;
  if (finalz < Zmin) finalz = Zmin;
  
  // -------------------------------------------------------------------- send X/Y/Z values to IotHAL
  IotModule.genericWrite16(pinX, k_Center + int(finalx * k_MinMax));
  IotModule.genericWrite16(pinY, k_Center + int(finaly * k_MinMax));
  IotModule.genericWrite16(pinZ, k_Center + int(finalz * k_MinMax));

  // -------------------------------------------------------------------- set zero if pinCalibrate = 0
  if (!digitalRead(pinCalibrate))
  { 
    x = 0;
    y = 0;
    z = 0;  
    calx = euler.x();
    caly = euler.y();
    calz = euler.z();
  }
  
  // -------------------------------------------------------------------- loop time test (total = 1.5 to 2.5 mS) (248 Hz)
  digitalWrite(pinTest, !digitalRead(pinTest));
}


// ============================================================================
//  FUNCTIONS - TEMPERATURE
// ============================================================================

void displayTemperature()
{
  int8_t temp = bno.getTemp();
  Serial.print("Current Temperature: ");
  Serial.print(temp);
  Serial.println(" °C\n");
}

// ============================================================================
//  FUNCTIONS - CALIBRATION AND OFFSETS
// ============================================================================
void displayCalibrationOffsets(const adafruit_bno055_offsets_t &calibData, String title)
{
    Serial.println("");
    Serial.println("--------------- " + title);
    
    Serial.print("Accelerometer: ");
    Serial.print(calibData.accel_offset_x); Serial.print(" ");
    Serial.print(calibData.accel_offset_y); Serial.print(" ");
    Serial.print(calibData.accel_offset_z); Serial.print(" ");

    Serial.print("\nGyro: ");
    Serial.print(calibData.gyro_offset_x); Serial.print(" ");
    Serial.print(calibData.gyro_offset_y); Serial.print(" ");
    Serial.print(calibData.gyro_offset_z); Serial.print(" ");

    Serial.print("\nMag: ");
    Serial.print(calibData.mag_offset_x); Serial.print(" ");
    Serial.print(calibData.mag_offset_y); Serial.print(" ");
    Serial.print(calibData.mag_offset_z); Serial.print(" ");

    Serial.print("\nAccel Radius: ");
    Serial.print(calibData.accel_radius);

    Serial.print("\nMag Radius: ");
    Serial.println(calibData.mag_radius);

    Serial.println("----------------------------------------\n");
}

void readCalibrationFromEeprom()
{
  int eeAddress = 0;
  adafruit_bno055_offsets_t calib;
  EEPROM.get(eeAddress, calib);
  bno.setSensorOffsets(calib);
  displayCalibrationOffsets(calib, "Calibration from EEPROM");
}

void writeCalibrationToEeprom()
{
  int eeAddress = 0;
  adafruit_bno055_offsets_t calib;
  bno.getSensorOffsets(calib);
  EEPROM.put(eeAddress, calib);
  displayCalibrationOffsets(calib, "Calibration to EEPROM");
}

void waitCalibration()
{
  Serial.println("Move sensor slightly to calibrate magnetometers");
  while (1)
  {
    displayCalibrationStatus();
    if (bno.isFullyCalibrated()) break;
    digitalWrite(pinLed, !digitalRead(pinLed));
    delay(200);
  }
  digitalWrite(pinLed, 0);
}

void displayCalibrationStatus()
{
  uint8_t system, gyro, accel, mag = 0;
  bno.getCalibration(&system, &gyro, &accel, &mag);
  Serial.print("CALIBRATION: Sys=");
  Serial.print(system, DEC);
  Serial.print(" Gyro=");
  Serial.print(gyro, DEC);
  Serial.print(" Accel=");
  Serial.print(accel, DEC);
  Serial.print(" Mag=");
  Serial.println(mag, DEC);
}
