//================================================================================================
//                                  cIotModuleAdc header
//================================================================================================

#define AD7124_PIN_NUM_MOSI (2)  //adc24
#define AD7124_PIN_NUM_MISO (4)  //adc24_din
#define AD7124_PIN_NUM_CLK  (0)  //adc24_dout diagnostic ramp

#define AD7124_STATUS_REGISTER (0)
#define AD7124_ADC_CONTROL_REGISTER (1)
#define AD7124_DATA_REGISTER (2)
#define AD7124_ID_REGISTER (5)
#define AD7124_ERROR_REGISTER (6)

// COMMS REGISTER READ bit
#define AD7124_COMMS_WRITE_FLAG             (1 << 6)

// AD7124_ADC_CONTROL_REGISTER bits
#define AD7124_ADC_CTRL_REG_DOUT_RDY_DEL    (1 << 12)
#define AD7124_ADC_CTRL_REG_CONT_READ       (1 << 11)
#define AD7124_ADC_CTRL_REG_DATA_STATUS     (1 << 10)
#define AD7124_ADC_CTRL_REG_CS_EN           (1 << 9)
#define AD7124_ADC_CTRL_REG_REF_EN          (1 << 8)
#define AD7124_ADC_CTRL_REG_POWER_MODE(x)   (((x) & 0x3) << 6)
#define AD7124_ADC_CTRL_REG_MODE(x)         (((x) & 0xF) << 2)
#define AD7124_ADC_CTRL_REG_CLK_SEL(x)      (((x) & 0x3) << 0)

// AD7124_STATUS_REGISTER bits
#define AD7124_STATUS_REGISTER_ERROR_FLAG   (1 << 6)
#define AD7124_STATUS_REGISTER_POR_FLAG     (1 << 4)

// AD7124_ERROR_REGISTER bits
#define AD7124_ERROR_REGISTER_SPI_IGNORE_ERR (1 << 6)

// --- Configuration Registers 0-7 bits
#define AD7124_CFG_REG_BIPOLAR     (1 << 11)
#define AD7124_CFG_REG_BURNOUT(x)  (((x) & 0x3) << 9)
#define AD7124_CFG_REG_REF_BUFP    (1 << 8)
#define AD7124_CFG_REG_REF_BUFM    (1 << 7)
#define AD7124_CFG_REG_AIN_BUFP    (1 << 6)
#define AD7124_CFG_REG_AINN_BUFM   (1 << 5)
#define AD7124_CFG_REG_REF_SEL(x)  ((x) & 0x3) << 3
#define AD7124_CFG_REG_PGA(x)      (((x) & 0x7) << 0)

// --- Channel Registers 0-15 bits
#define AD7124_CH_MAP_REG_CH_ENABLE    (1 << 15)

// --- Filter Register 0-7 bits
#define AD7124_FILT_REG_FILTER(x)         (((uint32_t)(x) & 0x7) << 21)
#define AD7124_FILT_REG_REJ60             ((uint32_t)1 << 20)
#define AD7124_FILT_REG_POST_FILTER(x)    (((uint32_t)(x) & 0x7) << 17)
#define AD7124_FILT_REG_SINGLE_CYCLE      ((uint32_t)1 << 16)
#define AD7124_FILT_REG_FS(x)             (((uint32_t)(x) & 0x7FF) << 0)

// --- Registers
#define AD7124_IOCon2    (4)
#define AD7124_Channel_0 (9)
#define AD7124_Config_0  (0x19)
#define AD7124_Filter_0  (0x21)

class cIotModuleAdc
{
  public:
    static const uint8_t AD7124_MOSI = 0;
    static const uint8_t AD7124_MISO = 1;
    static const uint8_t AD7124_CLK = 2;

  private:
    gpio_num_t SPI_MOSI_Name;
    gpio_num_t SPI_MISO_Name;
    gpio_num_t SPI_CLK_Name;
    gpio_pull_mode_t SPI_MOSI_ResetMode;
    gpio_pull_mode_t SPI_MISO_ResetMode;
    gpio_pull_mode_t SPI_CLK_ResetMode;
    
    uint16_t _Sps;
    bool _MultipleSetups;
    bool _FiltersSingleCycle;
    uint8_t _PreFilterType;
    uint8_t _PostFilterType;
    uint16_t _ActiveChannels;
    uint16_t _BiasedChannels;
    uint16_t _DiffChannels;
    uint16_t _PseudoChannels;
    uint8_t _ChannelsGains0;
    uint8_t _ChannelsGains1;
    uint8_t _ChannelsGains2;

    spi_device_handle_t hSpi;
    uint8_t ISR_Counter;
    uint32_t ISR_Channel_uint32_Value[16];
    bool SpiActive;
    bool ISR_Active;
    bool MasterPinConfigured;

    bool ContinuousConversion_ContRead;

    void ISR_Attach(void);
    void ISR_Detach(void);
    void SPI_Attach(const uint8_t PinNumClk, const uint8_t PinNumMiso, const uint8_t PinNumMosi, const uint32_t ClockSpeedHz);
    void SPI_Detach(void);
    void Reset(void);
    void Init_ContinuousConversion(void);
    void Init_ContinuousConversion_ContRead(void);
    void Init_Idle(void);
    uint8_t ReadStatusRegister(void);
    uint8_t ReadID_Register(void);
    void ProgramAllRegisters(void);
    void SetFixedParam(void);
    void WriteReg16(const uint8_t reg, const uint16_t data16);
    void WriteReg24(const uint8_t reg, const uint32_t data24);
    uint8_t ReadReg8(const uint8_t reg);
    uint32_t ReadReg24(const uint8_t reg);
    static void SPI_ISR_Handler(void *arg);
  public:

    cIotModuleAdc();
    uint16_t GetActiveChannels(void);
    bool GetMasterPinConfigured(void);
    void ConfigureMasterPin(void);
    void ShutDown(void);
    void SetSPI_Control_Pin(uint8_t Type, gpio_num_t Name, gpio_pull_mode_t ResetMode);
    void WaitRegistersWritableBySPI(void);
    uint8_t GetISR_Counter();
    uint32_t GetISR_Channel_uint32_Value(const uint8_t Index);
    void SetParam(const uint16_t Sps, const bool MultipleSetups, const bool FiltersSingleCycle, const uint8_t PreFilterType,
                  const uint8_t PostFilterType, const uint16_t ActiveChannels, const uint16_t BiasedChannels, const uint16_t DiffChannels,
                  const uint16_t PseudoChannels, const uint8_t ChannelsGains0, const uint8_t ChannelsGains1, const uint8_t ChannelsGains2);
    void Initialize(void);
    void PreFuota(void);

};
