
// ==================================================================================
//  This program reads a MPU-6050 (ACC GYRO) and sends AccX, AccY, AccZ to IotHAL
//  Values are sent to IotHAL as 16 bits from 0 to 65535
// ==================================================================================

// ==================================================================================
//   ESP32 to MPU-6050
// ==================================================================================
//   ESP32 Pin      MPU-6050 
//  ---------------------------------------------------------------------------------
//     GND          GND 
//     +5V          VIN
//     27           SDA        
//     14           SCL            
// ----------------------------------------------------------------------------------   

// ==================================================================================
//   ESP32 TO IOT-HAL
// ==================================================================================
//   ESP32 Pin     Signal type                        IotHal Pin configuration 
//  ---------------------------------------------------------------------------------
//      36          X (acceleration right-left)         GenericIn16
//      39          Y (acceleration front-rear)         GenericIn16
//      25 (or 34)  Z (acceleration up-down)            GenericIn16
// ----------------------------------------------------------------------------------     

#include "IotModule\IotModule.h"
#include "Utility/ThereminoFilters.h"
#include <Wire.h>

// ===================================
//  IotModule object, do not remove
// ===================================
cIotModule IotModule;
// ===================================

// -------------------------------------------------------- eventually modify pins
byte pinX = 36;
byte pinY = 39;
byte pinZ = 25;    // 25 on Pico - 34 on Wroom
byte pinSDA = 27;
byte pinSCL = 14;
byte pinTest = 21;

// ------------------------------------------------------------------------------
//  Filters to transform accelerometers to velocimeters (see documentation)
// ------------------------------------------------------------------------------
//  With 4.5 Hz the resulting signal is similar to a 4.5 Hz Geophone 
//  You could change HP and LP freq. to get different behaviours
//  Or you could cmment the following USE_FILTERS to get unfiltered data
// ------------------------------------------------------------------------------
#define USE_FILTERS
const float HiPassFreq  = 4.5;
const float LowPassFreq = 4.5;
// -------------------------------------------------------- filters declarations 
Filter hipass1(HiPassFreq, true);
Filter hipass2(HiPassFreq, true);
Filter hipass3(HiPassFreq, true);
Filter lopass1(LowPassFreq, false);
Filter lopass2(LowPassFreq, false);
Filter lopass3(LowPassFreq, false);

// -------------------------------------------------------- do not modify this part
int MPU = 0x68;  // I2C address of the MPU-6050
float x, y, z;

void setup(void)
{
  // =========================================================
  //  INIT SERIAL
  // =========================================================
  Serial.begin(115200);

  // =========================================================
  //  IotModule object initialization, do not remove
  // =========================================================
  IotModule.initialize();

  // =========================================================
  //  pin mode for PIN TEST
  // =========================================================
  pinMode(pinTest, OUTPUT);

  // =========================================================
  //  Init I2C communication  (SDA / SCL / FREQUENCY)
  // =========================================================
  Wire.begin(pinSDA, pinSCL);

  // ========================================================
  //  CONFIG THE MPU-6050
  // ========================================================
  // -------------------------------------------------------- wake up the MPU-6050
  Wire.beginTransmission(MPU);
  Wire.write(0x6B);  // PWR_MGMT_1 register
  Wire.write(0);     // set to zero (wakes up the MPU-6050)
  Wire.endTransmission(true);
  // -------------------------------------------------------- G scale config
  Wire.beginTransmission(MPU);
  Wire.write(0x1C);         // ACCEL_CONFIG register
  Wire.write(0b00000000); // +/-2G
  //Wire.write(0b00001000); // +/-4G
  //Wire.write(0b00010000); // +/-8G
  //Wire.write(0b00011000); // +/-16G
  Wire.endTransmission(true);
  // -------------------------------------------------------- bandpass config (decrease to reduce the noise)
  Wire.beginTransmission(MPU);
  Wire.write(0x1A);         // CONFIG register
  //Wire.write(0b00000000); // 260 Hz
  //Wire.write(0b00001000); // 184 Hz
  Wire.write(0b00010000); //  94 Hz
  //Wire.write(0b00011000); //  44 Hz
  //Wire.write(0b00100000); //  21 Hz
  //Wire.write(0b00101000); //  10 Hz
  //Wire.write(0b00110000); //   5 Hz
  Wire.endTransmission(true);
  // -------------------------------------------------------- 
}

void loop(void)
{
  // -------------------------------------------------------- starting with register 0x3B (ACCEL_XOUT_H)
  Wire.beginTransmission(MPU);
  Wire.write(0x3B);
  Wire.endTransmission(false);
  // -------------------------------------------------------- request a total of 6 registers (1 to 1.5 mS)
  Wire.requestFrom((uint16_t) MPU, (uint8_t) 6, true);
  int16_t ix, iy, iz;
  ix = (((int16_t)Wire.read()) << 8) | ((int16_t)Wire.read());
  iy = (((int16_t)Wire.read()) << 8) | ((int16_t)Wire.read());
  iz = (((int16_t)Wire.read()) << 8) | ((int16_t)Wire.read());

  // -------------------------------------------------------- scale the G sensitivity
  const float k = 1.0 / 32768;  // normalized from -1G to +1G
  x = k * (double)ix;
  y = k * (double)iy;
  z = k * (double)iz;
  
  // -------------------------------------------------------- HiPass and LowPass filters
  #ifdef USE_FILTERS
    x = hipass1.run(x);
    x = lopass1.run(x);
    y = hipass2.run(y);
    y = lopass2.run(y);
    z = hipass3.run(z);
    z = lopass3.run(z);
  #endif

  // -------------------------------------------------------- send X/Y/Z values to IotHAL as float numbers
  //                      normalized from 0 to 1000
  IotModule.genericWriteFloat(pinX, 500 + x * 1000);
  IotModule.genericWriteFloat(pinY, 500 + y * 1000);
  IotModule.genericWriteFloat(pinZ, 500 + z * 1000);
  
  // -------------------------------------------------------- loop time test (total = 1.5 to 2.5 mS) (248 Hz)
  //digitalWrite(pinTest, !digitalRead(pinTest));
}
